#include <stdlib.h>
#include <string.h>

#include "MibCompiler.h"
#include "symbol.h"
#include "AssignOID.h"


extern int bResolveOIDsNeeded,nCheckOnly;


void AddOIDSymbol(char *pName,LIST *pOID,int nLineNo)
 {
    SYMBOL *pSym;
    /*
     * Make sure that the symbol isn't already known to us.
     */
    if( pSym = FindSymbol(pName,OIDTABLE) )
             {
                /*
                 * Yes. See if it is one of the names we added. If so,
		 * we don't do anything. Otherwise we complain.
                 */
                if( pSym->DefinedLineNo )
                    RestoreErrorMsg("Line %d : Symbol defintion of \"%s\" is already defined on line %d\n",
                                        nLineNo,pName,pSym->DefinedLineNo);
             }
        else {
                /*
                 * It isn't known to us, so add it to the system.
                 */
                pSym = (SYMBOL *) malloc(sizeof(SYMBOL));
                if( pSym )
                         {
                            /*
                             * Okay. So fill in the details.
                             */
                            pSym->DefinedLineNo = nLineNo;
                            pSym->SymbolName = strdup(pName);
                            pSym->SymbolValue = pOID;

                            /*
                             * Add the symbol to the system.
                             */
                            AddToSymbolTable(pSym,OIDTABLE);
                         }
                    else {
                            /*
                             * Out of memory!
                             */
                            RestoreErrorMsg("Out of memory\n");
                            exit(1);
                         }
             }
 }


static LIST *GraftLists(LIST *pS,LIST *pAdd)
 {
    LIST *pNext = (LIST *)pS->Next;
    LIST *pFirst = pS;
    for( ; pAdd; )
        {
            /*
             * Copy the pAdd element over the pS element.
             */
            memcpy((char *)pS,(char *)pAdd,sizeof(LIST));

            /*
             * Advance to the next to add in.
             */
            pAdd = (LIST *)pAdd->Next;

            /*
             * If we have another to add, we need to allocate space 
             * for it.
             */
            if( pAdd )
                {
                    pS->Next = (struct list *) malloc(sizeof(LIST));
                    if( pS->Next )
                             {
                                /*
                                 * Ok. 
                                 */
                                pS = (LIST *)pS->Next;
                             }
                        else {
                                /*
                                 * Too bad, we are out of memory.
                                 */
                                RestoreErrorMsg("Out of memory\n");
                                return((LIST *)0);
                             }
                }
        }

    /*
     * Make the last one point to the rest of the list.
     */
    pS->Next = (struct list *)pNext;

    /*
     * Return a pointer to the insertion point.
     */
    return(pS);
 }


/*
 * Convert a Name/Number oid list to just numbers, such as 1.3.6.
 */
LIST *ResolveSpecificOID(LIST *pL,char *pName,int bPrintMsg)
 {
    SYMBOL *pSym;
    LIST *pT,*pFirst = pL;
    /*
     * For each element in the list, make sure we have a number for it.
     */
    for( ; pL; pL = (LIST *)pL->Next )
        {
            /*
             * If this is a name only entry we need to resolve it.  
             */
            if( pL->Type == NAMEONLY )
                {
                    /*
                     * First, make sure that we have a symbol defined.
                     */
                    pSym = FindSymbol(pL->Name,OIDTABLE);
                    if(pSym)
                             {
                                /*
                                 * Ok. Resolve this specific symbol now.
                                 */
                                pT = ResolveSpecificOID(pSym->SymbolValue,pL->Name,bPrintMsg);
                                if( pT )
                                         {
                                            /*
                                             * Okay, we need to graft the
                                             * two lists together.
                                             */
                                            GraftLists(pL,pT);
                                         }
                                    else {
                                            /*
                                             * Too bad, we couldn't resolve
                                             * it.
                                             */
                                            return((LIST *)0);
                                         }
                             }
                        else {
                                /*
                                 * Too bad!
                                 */
                                if( bPrintMsg )
                                    {
                                        if( pName )
                                             RestoreErrorMsg("While trying to resolve object id \"%s\", the symbol \"%s\" couldn't be found\n",pName,pL->Name);
                                        else RestoreErrorMsg("While trying to resolve an object id, the symbol %s couldn't be found\n",pL->Name);
                                    }
                                return((LIST *)0);
                             }
                }
        }

    /*
     * Return the list head.
     */
    return(pFirst);
 }


/*
 * See if all the OIDs that where defined are complete -- they can
 * be resolved into numbers.
 */
int AllOIDsResolved(void)
 {
    SYMBOL *pSym;
    int nRc = 1,nTemp;
    /*
     * For each symbol in the table, do...
     */
    for( pSym = 0; pSym = GetNextSymbol(pSym,OIDTABLE); )
        {
            nTemp = (ResolveSpecificOID(pSym->SymbolValue,pSym->SymbolName,1) != (LIST *)0);
            if( nRc ) nRc = nTemp;
        }
    return( nRc );
 }


/*
 * Dump everything in the OID symbol table.
 */
void DumpAllOIDs(void)
 {
    SYMBOL *pSym;
    /*
     * For each symbol in the table, do...
     */
    for( pSym = 0; pSym = GetNextSymbol(pSym,OIDTABLE); )
        {
            printf("Symbol: %s\n",pSym->SymbolName);
            PrintList(pSym->SymbolValue,0);printf("\n");
        }
 }


/*
 * pName	name of the OID definition.
 * pOID		pointer to linked list defining the OID (watch out
 *			the linked list can have both names and
 *			numbers in it)
 * nLineNo	definition of OID starts on nLineNo in MIB file.
 */
void ProcessOID(char *pName,LIST *pOID,int nLineNo)
 {
    LIST *pNew;
    /*
     * See if we can resolve the OID to numbers only at this time.
     */
    *pNew = ResolveSpecificOID(pOID,pName,0);

    /*
     * Use the resolved address, if we can. This will save us a lot of
     * time in later steps.
     */
    if( pNew )
           pOID = pNew;
      else bResolveOIDsNeeded++;

    /*
     * Add the symbol to the table.
     */
    AddOIDSymbol(pName,pOID,nLineNo);
 }



void AddStandardOIDS(void)
 { 
        LIST *pList;
	/*
         * The first "standard" names are iso, ccitt and joint-iso-ccitt.
         */
        pList = newListElement(NUMBERONLY,(char *)0,0);
        AddOIDSymbol("ccitt",pList,0);
        pList = newListElement(NUMBERONLY,(char *)0,1);
        AddOIDSymbol("iso",pList,0);
        pList = newListElement(NUMBERONLY,(char *)0,2);
        AddOIDSymbol("joint-iso-ccitt",pList,0);

        /*
         * Now add the things from RFC-1155.
         * Start with internet.
         */
        pList = newListElement(NAMEONLY,"iso",0);
        pList = LinkToList(pList,newListElement(NAMEANDNUMBER,"org",3));
        pList = LinkToList(pList,newListElement(NAMEANDNUMBER,"dod",6));
        pList = LinkToList(pList,newListElement(NUMBERONLY,(char *)0,1));
        ProcessOID("internet",pList,0);

        /*
         * Add directory.
         */
        pList = newListElement(NAMEONLY,"internet",0);
        pList = LinkToList(pList,newListElement(NUMBERONLY,(char *)0,1));
        ProcessOID("directory",pList,0);

        /*
         * The 'mgmt' thing.
         */
        pList = newListElement(NAMEONLY,"internet",0);
        pList = LinkToList(pList,newListElement(NUMBERONLY,(char *)0,2));
        ProcessOID("mgmt",pList,0);

        /*
         * The 'experimental' thing.
         */
        pList = newListElement(NAMEONLY,"internet",0);
        pList = LinkToList(pList,newListElement(NUMBERONLY,(char *)0,3));
        ProcessOID("experimental",pList,0);

        /*
         * The 'private' thing.
         */
        pList = newListElement(NAMEONLY,"internet",0);
        pList = LinkToList(pList,newListElement(NUMBERONLY,(char *)0,4));
        ProcessOID("private",pList,0);

        /*
         * The 'enterprises' thing.
         */
        pList = newListElement(NAMEONLY,"private",0);
        pList = LinkToList(pList,newListElement(NUMBERONLY,(char *)0,1));
        ProcessOID("enterprises",pList,0);
 }
