#include <stdlib.h>

#include "MibCompiler.h"

extern char yytext[];
extern int yylineno;

int yyErrorCnt = {0};


yyerror(char *pStr)
 {
    RestoreErrorMsg("Line: %d  %s (Last token %s)\n",yylineno,pStr,yytext);
    yyErrorCnt++;
 }


TYPE *newTypeElement(TYPEofTYPE type,char *pName)
 {
    TYPE *pRc;
    pRc = (TYPE *) malloc(sizeof(TYPE));
    if( pRc == (TYPE *)0 )
        {
            RestoreErrorMsg("Out of memory\n");
            return((TYPE *)0);
        }
    memset(pRc,0,sizeof(TYPE));
    pRc->Type = type;
    switch(type)
        {
            case TYPENAME : 
                            pRc->Data.TypeName = pName;
                            break;
            case SEQTYPE :
            case SEQOFTYPE:
                            pRc->Data.TypeList = pName;
                            break;
            case ENUMINTEGER :
                            pRc->Data.EnumList = (LIST *)pName;
                            break;
            case STRINGTYPE:
            case OCTETTYPE:
            case RANGEINTEGER :
                            memcpy(&pRc->Data.Range,(PAIR *)pName,sizeof(PAIR));
                            break;
        }
    return(pRc);
 }


LIST *newListElement(ENTRYTYPE type,char *pName,int nValue)
 {
    LIST *pRc;
    pRc = (LIST *) malloc(sizeof(LIST));
    if( pRc == (LIST *)0 )
        {
            RestoreErrorMsg("Out of memory\n");
            return((LIST *)0);
        }
    memset(pRc,0,sizeof(LIST));
    pRc->Type = type;
    pRc->Name = pName;
    pRc->Number = nValue;
    return(pRc);
 }


LIST *LinkToList(LIST *p1,LIST *p2)
 {
    LIST *pTemp = p1;
    while( pTemp->Next ) pTemp = pTemp->Next;
    pTemp->Next = p2;
    return(p1);
 }


TYPE *LinkToType(TYPE *p1,TYPE *p2)
 {
    TYPE *pTemp = p1;
    while( pTemp->Next ) pTemp = pTemp->Next;
    pTemp->Next = p2;
    return(p1);
 }


char *OIDListToString(LIST *pL)
 {
    static char buffer[128];
    char *pStr;
    for( pStr = buffer; pL; pL = (LIST *)pL->Next )
        {
            sprintf(pStr,"%d",pL->Number);
            pStr += strlen(pStr);
            if( pL->Next ) *pStr++ = '.';
        }
    return(buffer);
 }
